;;; Hoot read-eval-print loop
;;; Copyright (C) 2026 David Thompson <dave@spritely.institute>
;;;
;;; Licensed under the Apache License, Version 2.0 (the "License");
;;; you may not use this file except in compliance with the License.
;;; You may obtain a copy of the License at
;;;
;;;    http://www.apache.org/licenses/LICENSE-2.0
;;;
;;; Unless required by applicable law or agreed to in writing, software
;;; distributed under the License is distributed on an "AS IS" BASIS,
;;; WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
;;; See the License for the specific language governing permissions and
;;; limitations under the License.

(use-modules (fibers promises)
             (fibers streams)
             (hoot hackable)
             (hoot repl))

(lambda (resolved rejected . load-path)
  (call-with-async-result
   resolved rejected
   (lambda ()
     (parameterize ((current-input-port (standard-input-stream))
                    (current-output-port (standard-output-stream)))
       (for-each add-to-load-path! load-path)
       (spawn-repl)))))
