;;; Guile weak vectors
;;; Copyright (C) 2024 Vivianne Langdon <puttabutta@gmail.com>
;;;
;;; Licensed under the Apache License, Version 2.0 (the "License");
;;; you may not use this file except in compliance with the License.
;;; You may obtain a copy of the License at
;;;
;;;    http://www.apache.org/licenses/LICENSE-2.0
;;;
;;; Unless required by applicable law or agreed to in writing, software
;;; distributed under the License is distributed on an "AS IS" BASIS,
;;; WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
;;; See the License for the specific language governing permissions and
;;; limitations under the License.

;;; Commentary:
;;;
;;; Guile weak vectors module.
;;;
;;; Code:

(define-module (ice-9 weak-vector)
  #:use-module (hoot weak-vectors)
  #:re-export (make-weak-vector
               weak-vector
               list->weak-vector
               weak-vector?
               weak-vector-ref
               weak-vector-set!))
